##############################################################################
#
# License
# -------
#
# Copyright (c) Gerard Hale. All right reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Gerard Hale requests that attribution be given to Zope
#    in any manner possible. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
#
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
#
#      "This product includes software developed by Gerard Hale
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
# 5. Names associated with Gerard Hale must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Gerard Hale.
#
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
#
#      "This product includes software developed by Gerard Hale
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official software releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.
#
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY Gerard Hale ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL Gerard Hale OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
#
# This software consists of contributions made by Gerard Hale and
# many other individuals.
#
##############################################################################
"""Z Session product """
__version__='$Revision: 0.1.0'

import Globals
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from string import join, strip, split, lower
from Acquisition import Implicit, Explicit
from OFS.Document import Document
from urllib import quote, unquote
from cPickle import loads, dumps
import Acquisition
import OFS.SimpleItem
import AccessControl.Role
import binascii, base64, string
import time, threading, os, DateTime

"""The global Session Dictionary."""
SessionDict     = {}
"""The global Application Dictionary."""
ApplicationDict = {}
"""The global Application Locking Queue."""
ApplicationLockQueue = []

AppLockTimeout = 1000

PRODUCT_DIR = split(__name__, '.')[-2]

class ZSessionObj(Explicit):
    """ A ZSession Object. Installed into the REQUEST object as 
    REQUEST.SESSION (like REQUEST.RESPONSE)
    """

    __allow_access_to_unprotected_subobjects__ = 1

    ZSession = Acquisition.Acquired

    def __init__(self, sessionName):
        """Initialise ZSessionObj"""
        self.sName = sessionName
        if SessionDict.has_key(sessionName):
            self.cachedict = SessionDict[sessionName]
        else:
            self.cachedict = {}
        self.cachedict['ZSTimeStamp'] = int(time.time())

    def getName(self):
        """return name of session"""
        return self.sName

    def __getitem__(self, name):
        """get value of key"""
        if self.cachedict.has_key(name):
            return self.cachedict[name]
        else:
            return ""

    def __setitem__(self, name, value):
        """set value of key"""
        self.cachedict[name] = value
        SessionDict[self.getName()] = self.cachedict
        return None

    set = __setitem__

    def __delitem__(self, name):
        """delete key"""
        del self.cachedict[name]
        SessionDict[self.getName()] = self.cachedict
        return None

    delete = __delitem__

    def __len__(self):
        """get length of ZSession minus ZSTimeStamp"""
        return len(self.cachedict)-1

    len = __len__

    def keys(self):
        """get ZSession keys"""
        l = []
        for r in self.cachedict.keys():
            if(r == 'ZSTimeStamp'):
                pass
            else:
                l.append(r)
        return l

    def values(self):
        """get ZSession values"""
        values = []
        for n, v in self.cachedict.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    values.append(v)
            except:
                values.append(v)
        return values

    def items(self):
        """get ZSession key value pairs"""
        items = []
        for n, v in self.cachedict.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    items.append((n,v))
            except:
                items.append((n,v))
        return items

    def has_key(self, name):
        """check if ZSession has key"""
        if self.cachedict.has_key(name):
            return 1
        else:
            return 0

    def __str__(self):
        """get ZSession key value pairs"""
        l = ["ZSession Variables:\n"]
        for n,v in self.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    l.append("%s: %s"%(n,str(v)))
            except:
                l.append("%s: %s"%(n,str(v)))
        return join(l, '\n')

    def clear(self):
        """clear ZSession keys"""
        for key in self.cachedict.keys():
            if(key == 'ZSTimeStamp'):
                pass
            else:
                del self.cachedict[key]
        return None

    def copy(self):
        """can not copy a ZSession"""
        raise RuntimeError, "copy() unimplemented for ZSessions"

    def update(self, dict):
        """update ZSession key value pairs"""
        for k, v in dict.items():
            self[k] = v

    def get(self, key, failobj=None):
        """get ZSession key value"""
        if self.has_key(key): return self[key]
        else: return failobj

class ZApplicationObj(Explicit):
    """ A ZApplication Object. Installed into the REQUEST object as 
    REQUEST.APPLICATION (like REQUEST.RESPONSE)
    """

    __allow_access_to_unprotected_subobjects__ = 1

    ZSession = Acquisition.Acquired

    def __init__(self, applicationName, sessionName=None):
        """Initialise ZApplicationObj"""
        self.aName = applicationName
        self.AppLockTimeout = 0
        if sessionName is not None:
            self.sName = sessionName
        if ApplicationDict.has_key(applicationName):
            self.cachedict = ApplicationDict[applicationName]
        else:
            self.cachedict = {}
        self.lock()
        self.set('ZSTimeStamp',int(time.time()))
        self.unlock()

    def getName(self):
        """return name of application"""
        return self.aName

    def __getitem__(self, name):
        """get value of key"""
        if self.cachedict.has_key(name):
            return self.cachedict[name]
        else:
            return ""

    def __setitem__(self, name, value):
        """set value of key"""
        if(self.sName in ApplicationLockQueue):
            if(self.sName == ApplicationLockQueue[0]):
                self.cachedict[name] = value
                ApplicationDict[self.getName()] = self.cachedict
                self.AppLockTimeout = 0
                return "Value set"
            else:
                self.AppLockTimeout = self.AppLockTimeout + 1
                #call again
                if(self.AppLockTimeout < ApplicationSetTimeout):
                    self.set(name, value)
                else:
                    self.AppLockTimeout = 0
                    return "Value can not be set"
        else:
            return "Application Object is not Locked"

    set = __setitem__

    def __delitem__(self, name):
        """delete key"""
        if(self.sName in ApplicationLockQueue):
            if(self.sName == ApplicationLockQueue[0]):
                if self.cachedict.has_key(name):
                    del self.cachedict[name]
                    ApplicationDict[self.getName()] = self.cachedict
                self.AppLockTimeout = 0
                return "Key deleted"
            else:
                self.AppLockTimeout = self.AppLockTimeout + 1
                #call again
                if(self.AppLockTimeout < ApplicationDelTimeout):
                    self.delete(name)
                else:
                    self.AppLockTimeout = 0
                    return "Key can not be deleted"
        else:
            return "Application Object is not Locked"

    delete = __delitem__

    def __len__(self):
        """get length of ZApplication minus ZSTimeStamp"""
        return len(self.cachedict)-1

    len = __len__

    def keys(self):
        """get ZApplication keys"""
        l = []
        for r in self.cachedict.keys():
            if(r == 'ZSTimeStamp'):
                pass
            else:
                l.append(r)
        return l

    def values(self):
        """get ZApplication values"""
        values = []
        for n, v in self.cachedict.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    values.append(v)
            except:
                values.append(v)
        return values

    def items(self):
        """get ZApplication key value pairs"""
        items = []
        for n, v in self.cachedict.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    items.append((n,v))
            except:
                items.append((n,v))
        return items

    def has_key(self, name):
        """check if ZApplication has key"""
        if self.cachedict.has_key(name):
            return 1
        else:
            return 0

    def __str__(self):
        """get ZApplication key value pairs"""
        l = ["ZApplication Variables:\n"]
        for n,v in self.items():
            try:
                if(n == 'ZSTimeStamp'):
                    pass
                else:
                    l.append("%s: %s"%(n,str(v)))
            except:
                l.append("%s: %s"%(n,str(v)))
        return join(l, '\n')

    def clear(self):
        """clear ZApplication keys"""
        if(self.sName in ApplicationLockQueue):
            if(self.sName == ApplicationLockQueue[0]):
                for key in self.cachedict.keys():
                    if(key == 'ZSTimeStamp'):
                        pass
                    else:
                        del self.cachedict[key]
                self.AppLockTimeout = 0
                return "ZApplication cleared"
            else:
                self.AppLockTimeout = self.AppLockTimeout + 1
                #call again
                if(self.AppLockTimeout < ApplicationDelTimeout):
                    self.clear()
                else:
                    self.AppLockTimeout = 0
                    return "ZApplication can not be cleared"
        else:
            return "Application Object is not Locked"

    def copy(self):
        """can not copy a ZApplication"""
        raise RuntimeError, "copy() unimplemented for ZApplications"

    def update(self, dict):
        """update ZApplication key value pairs"""
        for k, v in dict.items():
            self[k] = v

    def get(self, key, failobj=None):
        """get ZApplication key value"""
        if self.has_key(key): return self[key]
        else: return failobj

    def lock(self):
        if self.sName is None:
            return 'Application can not be locked'
        else:
            ApplicationLockQueue.append(self.sName)
            return 'Application locked'

    def unlock(self):
        if self.sName is None:
            return 'Application can not be unlocked'
        else:
            try:
                ApplicationLockQueue.remove(self.sName)
                return 'Application unlocked'
            except ValueError:
                #Session Name not in Queue
                return 'Application was not locked'


class ZSession(OFS.SimpleItem.Item,
                 Persistent,
                 Implicit,
                 AccessControl.Role.RoleManager
                 ):
    """ A ZSession. """

    meta_type = 'ZSession'
    id        = 'ZSession'
    title     = 'ZSession'
    icon      = 'misc_/ZSession/ZSession_icon'

    manage_options=(
        {'label':'Applications',     'action':'manage_applications',
         'help':(PRODUCT_DIR, 'Applications.html'),},
        {'label':'Sessions',     'action':'manage_sessions',
         'help':(PRODUCT_DIR, 'Sessions.html'),},
        {'label':'Properties',   'action':'manage_main',
         'help':(PRODUCT_DIR, 'Properties.html'),},
        {'label':'Process',      'action':'manage_process',
         'help':(PRODUCT_DIR, 'Process.html'),},
        {'label':'Security',     'action':'manage_access',
         'help':(PRODUCT_DIR, 'Security.html'),},
    )

    __ac_permissions__=(
        ('View management screens', ('manage_main',)),
        ('Change ZSessions',        ('manage_edit',)),
        ('Delete ZSessions',        ('manage_clearSessions',)),
        ('Delete ZSession Items',   ('manage_clearSessionKey',)),
        ('Delete ZApplications',    ('manage_clearApplications',)),
        ('Delete ZAplication Items',('manage_clearApplicationKey',)),
        ('View', ('__call__','')),
    )

    manage_main=HTMLFile('dtml/main', globals())
    manage_sessions=HTMLFile('dtml/manage_sessions', globals())
    manage_session_details=HTMLFile('dtml/manage_session_details', globals())
    manage_applications=HTMLFile('dtml/manage_applications', globals())
    manage_application_details=HTMLFile('dtml/manage_application_details', globals())
    manage_process=HTMLFile('dtml/manage_process', globals())

    def __init__(self, cookie_name, SessionTimeout, path='/'):
        """Initialise ZSession Object in Zope"""
        self.cookie_name = cookie_name
        self.SessionTimeout = SessionTimeout
        self.path = path
        self.lastRunTime = time.asctime(time.localtime(time.time()))
        self.sessionsCleared = []

    def manage_edit(self, cookie_name, path, SessionTimeout, REQUEST=None):
        """Change properties"""
        self.cookie_name = cookie_name
        self.SessionTimeout = SessionTimeout
        self.path = path
        try:
            ZSession.lastRunTime = ZSession.lastRunTime
            ZSession.sessionsCleared = ZSession.sessionsCleared
        except:
            pass

        if REQUEST:
            return MessageDialog(
                title   ='ZSession Changed',
                message ='ZSession properties have been updated',
                action  = REQUEST['URL1']+'/manage_main',
                target  ='manage_main')

    def __call__(self, applicationName=None, REQUEST=None):
        """
        Method called which checks or sets cookie and creates new
        ZSession Object and Application Object.
        """
        if REQUEST is None:
            try:
                REQUEST=self.REQUEST
            except:
                raise "Session Error", "can't find REQUEST, even in self!"
        response = REQUEST['RESPONSE']
        if REQUEST.cookies.has_key(self.cookie_name):
            sessionName = REQUEST.cookies[self.cookie_name]
        else:
            sessionName = self.newSession(REQUEST)
        self.buildSessionObj(REQUEST, sessionName)
        if applicationName is None:
            pass
        else:
            self.buildApplicationObj(REQUEST, applicationName, sessionName)
        blnThreadFound = 0
        counter = 0
        for t in threading.enumerate():
            if (t.getName() == "ZSession"):
                blnThreadFound = 1
                counter = counter + 1
        if(blnThreadFound==0):
            logProcess("__call__: Clear Sessions Thread has not started or exited!")
            cleanup=threading.Thread(target=self.clearSessionsProcess, name='ZSession', args=())
            cleanup.setDaemon(1) # so thread will terminate with parent process
            cleanup.start()
        """If dtml-var is used print Session Name"""
        return sessionName

    def buildSessionObj(self, REQUEST, sessionName):
        """Build the ZSessionObj for the individual ZSession"""
        s = ZSessionObj(sessionName)
        REQUEST['SESSION'] = s.__of__(self)

    def buildApplicationObj(self, REQUEST, applicationName, sessionName=None):
        """Build the ApplicationObj for the individual ZApplication"""
        s = ZApplicationObj(applicationName, sessionName)
        REQUEST['APPLICATION'] = s.__of__(self)

    def newSession(self, REQUEST, sessionName=None):
        " create a new session and set cookie. "
        response = REQUEST['RESPONSE']
        if sessionName is None:
            sessionName = self.makeSessionName(REQUEST)
            SessionDict[sessionName]={}
        response.setCookie(self.cookie_name, sessionName, path=self.path)
        REQUEST.cookies[self.cookie_name] = sessionName
        self.buildSessionObj(REQUEST, sessionName)
        return sessionName

    def newApplication(self, REQUEST, applicationName=None):
        " create a new application and set cookie. "
        if applicationName is not None:
            response = REQUEST['RESPONSE']
            ApplicationDict[applicationName] = {}
            response.setCookie(self.cookie_name, applicationName, path=self.path)
            REQUEST.cookies[self.cookie_name] = applicationName
            self.buildApplicationObj(REQUEST, applicationName)
        return applicationName

    def makeSessionName(self, REQUEST):
        """ create a new unique string to name the session. feel free to
            fiddle with this. """
        import md5, time, base64
        m = md5.new()
        m.update('A unique string is added here???')
        m.update(str(time.time()))
        m.update(str(REQUEST))
        return string.replace(base64.encodestring(m.digest())[:-3], '/', '$')

    def getSessionData(self, SessionName=None):
        """Get all the ZSession key value pairs created for the manage screens"""
        tmpDict = {}
        tmpSessionItemList = []
        tmpSessionList = []
        if SessionName:
            tmpDict = SessionDict[SessionName]
            for n,v in tmpDict.items():
                try:
                    if(n == 'ZSTimeStamp'):
                        pass
                    else:
                        tmpSessionItemList.append((n,v))
                except:
                    tmpSessionItemList.append((n,v))
            tmpSessionItemList.sort()
            sd = SessionName, tmpSessionItemList
            tmpSessionList.append(sd)
        else:
            for k in SessionDict.keys():
                tmpSessionItemList = []
                tmpDict = SessionDict[k]
                for n,v in tmpDict.items():
                    try:
                        if(n == 'ZSTimeStamp'):
                            pass
                        else:
                            tmpSessionItemList.append((n,v))
                    except:
                        tmpSessionItemList.append((n,v))
                tmpSessionItemList.sort()
                sd = k, tmpSessionItemList
                tmpSessionList.append(sd)
        return tmpSessionList

    def getApplicationData(self, ApplicationName=None):
        """Get all the ZApplication key value pairs created for the manage screens"""
        tmpDict = {}
        tmpApplicationItemList = []
        tmpApplicationList = []
        if ApplicationName:
            tmpDict = ApplicationDict[ApplicationName]
            for n,v in tmpDict.items():
                try:
                    if(n == 'ZSTimeStamp'):
                        pass
                    else:
                        tmpApplicationItemList.append((n,v))
                except:
                    tmpApplicationItemList.append((n,v))
            tmpApplicationItemList.sort()
            sd = ApplicationName, tmpApplicationItemList
            tmpApplicationList.append(sd)
        else:
            for k in ApplicationDict.keys():
                tmpApplicationItemList = []
                tmpDict = ApplicationDict[k]
                for n,v in tmpDict.items():
                    try:
                        if(n == 'ZSTimeStamp'):
                            pass
                        else:
                            tmpApplicationItemList.append((n,v))
                    except:
                        tmpApplicationItemList.append((n,v))
                tmpApplicationItemList.sort()
                sd = k, tmpApplicationItemList
                tmpApplicationList.append(sd)
        return tmpApplicationList

    def showAllSessions(self):
        """Get all the ZSessions created for the manage screens"""
        SessionList = []
        if(SessionDict=={}):
            return []
        else:
            for k in SessionDict.keys():
                dict = SessionDict[k]
                if dict.has_key('ZSTimeStamp'):
                    ts = dict['ZSTimeStamp']
                    SessionList.append([time.asctime(time.localtime(ts)),k])
            SessionList.sort()
        return SessionList

    def showAllApplications(self):
        """Get all the ZApplications created for the manage screens"""
        ApplicationList = []
        if(ApplicationDict=={}):
            return []
        else:
            for k in ApplicationDict.keys():
                dict = ApplicationDict[k]
                if dict.has_key('ZSTimeStamp'):
                    ts = dict['ZSTimeStamp']
                    ApplicationList.append([time.asctime(time.localtime(ts)),k])
            ApplicationList.sort()
        return ApplicationList

    def manage_clearSessions(self, SessionNames=[], REQUEST=None):
        """Clear Session Data from Session Dictionary
        The Session specified in 'SessionNames' gets deleted."""
        if type(SessionNames) is type(''): SessionNames=[SessionNames]
        if not SessionNames:
            return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_sessions',)
        for SessionName in SessionNames:
            if SessionDict.has_key(SessionName):
                del SessionDict[SessionName]
        if REQUEST is not None:
            return self.manage_sessions(self, REQUEST)

    def manage_clearApplications(self, ApplicationNames=[], REQUEST=None):
        """Clear Application Data from Application Dictionary
        The Application specified in 'ApplicationNames' gets deleted."""
        if type(ApplicationNames) is type(''): ApplicationNames=[ApplicationNames]
        if not ApplicationNames:
            return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_applications',)
        for ApplicationName in ApplicationNames:
            if ApplicationDict.has_key(ApplicationName):
                del ApplicationDict[ApplicationName]
        if REQUEST is not None:
            return self.manage_applications(self, REQUEST)

    def manage_clearSessionKey(self, SessionName, SessionKeys=None, REQUEST=None):
        """Manage Clear Session Keys"""
        if SessionKeys is not None:
            if SessionDict.has_key(SessionName):
                tmpDict = SessionDict[SessionName]
                for k in SessionKeys:
                    del tmpDict[k]
        if REQUEST is not None:
            return self.manage_session_details(self, REQUEST)

    def manage_clearApplicationKey(self, ApplicationName, ApplicationKeys=None, REQUEST=None):
        """Manage Clear Application Keys"""
        if ApplicationKeys is not None:
            if ApplicationDict.has_key(ApplicationName):
                tmpDict = ApplicationDict[ApplicationName]
                for k in ApplicationKeys:
                    del tmpDict[k]
        if REQUEST is not None:
            return self.manage_application_details(self, REQUEST)

    def clearSessionsProcess(self):
        """ clears old sessions that have timed out """
        sleepSeconds = 15
        while 1:
            sessionsClearedList = []
            if(SessionDict!={}):
                for k in SessionDict.keys():
                    dict = SessionDict[k]
                    if dict.has_key('ZSTimeStamp'):
                        SessionTime = dict['ZSTimeStamp'] + int(self.SessionTimeout)*60
                        if (SessionTime < int(time.time())):
                            del SessionDict[k]
                            sessionsClearedList.append(k)
            ZSession.lastRunTime = time.asctime(time.localtime(time.time()))
            ZSession.sessionsCleared = sessionsClearedList
            time.sleep(sleepSeconds)

    def showProcessLog(self):
        """
        Only one thread should ever be running, but if for some reason there are more
        than one they will be added to the list and show in the manage process screen.
        """
        threadList = []
        for t in threading.enumerate():
            if (t.getName() == "ZSession"):
                threadProperties = {}
                threadProperties['getName'] = t.getName()
                threadProperties['isAlive'] = t.isAlive()
                threadProperties['isDaemon'] = t.isDaemon()
                """This assumes there is only one thread."""
                try:
                    threadProperties['lastRunTime'] = ZSession.lastRunTime
                    threadProperties['sessionsCleared'] = ZSession.sessionsCleared
                except:
                    threadProperties['lastRunTime'] = 'Error: ZSession not updated yet'
                    threadProperties['sessionsCleared'] = 'Error: ZSession not updated yet'
                threadList.append(threadProperties)
        return threadList

    def ZSessionCount(self):
        return len(SessionDict)

    def ZApplicationCount(self):
        return len(ApplicationDict)

"""Global Functions"""

def logProcess(msg):
    """log process for the manage screens, may have trouble if Zope can not write to var directory"""
    LogFile = os.path.join(Globals.data_dir, 'ZSession.log')
    if(os.path.exists(LogFile)):
        try:
            f = open(LogFile, 'a')
            f.write(time.asctime(time.localtime(time.time())) + ' : ' + msg + '\n')
            f.close()
        except:
            """can not write to var directory.??"""
            pass
    else:
        """File does not exist yet"""
        try:
            f = open(LogFile, 'w')
            f.write(time.asctime(time.localtime(time.time())) + ' : ' + msg + '\n')
            f.close()
        except:
            """can not write to var directory.??"""
            pass
